import contextlib
import datetime
import time
from typing import Callable, Optional

STRFTIME_FORMAT = "%Y-%m-%d-%H-%M-%S"


def now() -> str:
    """Return current time in the format: %Y-%m-%d-%H-%M-%S.

    Returns:
        str: current time
    """
    STRFTIME_FORMAT = "%Y-%m-%d-%H-%M-%S"
    return datetime.datetime.now().strftime(STRFTIME_FORMAT)


class Timing(contextlib.ContextDecorator):
    """Timing Context Decorator.

    Args:
        prefix (str, optional): name of activity to time. Defaults to "".
        on_exit (Optional[Callable]): function to call on exit. Defaults to None.
        enabled (bool, optional): whether to enable timing. Defaults to True.
    """

    def __init__(self, prefix: str = "", on_exit: Optional[Callable] = None, enabled: bool = True) -> None:
        self.prefix = prefix
        self.on_exit = on_exit
        self.enabled = enabled

    def __enter__(self) -> None:
        """Start timing."""
        self.st = time.perf_counter_ns()

    def __exit__(self, *exc) -> None:
        """End timing and print elapsed time."""
        self.et = time.perf_counter_ns() - self.st
        if self.enabled:
            print(f"{self.prefix}: {self.et*1e-9:.2f} s" + (self.on_exit(self.et) if self.on_exit else ""))



from dataclasses import dataclass
from typing import Dict, List, Optional

import mlx.core as mx

from mlx_llm.model import create_model, create_tokenizer, quantize
from mlx_llm.prompt import create_prompt
from mlx_llm.utils.session import Session

from .utils import GO, STOP, WAIT, answer_status


@dataclass
class ChatSetup:
    """Dataclass for chat setup

    Raises:
        ValueError: Each element in history must contain a question and an answer.
    """

    system: str
    history: Optional[List[Dict[str, str]]] = None

    def session(self) -> Session:
        """Create session

        Raises:
            ValueError: Each element in history must contain a question and an answer.

        Returns:
            Session: session
        """
        for el in self.history:  # type: ignore
            if "question" not in el.keys() or "answer" not in el.keys():
                raise ValueError("Each element in history must contain a question and an answer.")
        if self.history is None:
            return Session()
        else:
            return Session(
                questions=[el["question"] for el in self.history], answers=[el["answer"] for el in self.history]
            )


class LLMChat:
    """Chat with LLM class

    Args:
        model_name (str): model name
        chat_setup (ChatSetup): chat setup
        max_tokens (int, optional): max tokens. Defaults to 1024.
        temperature (float, optional): temperature. Defaults to 0.1.
        quantized (bool, optional): quantized. Defaults to False.
        group_size (int, optional): group size. Defaults to 64.
        bits (int, optional): bits. Defaults to 8.
    """

    def __init__(
        self,
        model_name: str,
        chat_setup: ChatSetup,
        max_tokens: int = 1024,
        temperature: float = 0.1,
        quantized: bool = False,
        group_size: int = 64,
        bits: int = 8,
    ) -> None:
        self.model = create_model(model_name)
        self.tokenizer = create_tokenizer(model_name)
        self.prompt = create_prompt(model_name=model_name, system=chat_setup.system)
        self.session = chat_setup.session()
        self.max_tokens = max_tokens
        self.temperature = temperature

        if quantized:
            print(f"[INFO] Quantizing model {model_name} with group_size={group_size} and bits={bits}")
            self.model = quantize(self.model, group_size=group_size, bits=bits)

        self.model.eval()

    def start(self) -> None:
        """Start chat"""
        print("\n[INFO] Start chatting! Press 'q' to quit or 'r' to reset conversation.\n")

        print("******* CHAT *******")
        print(f"[SYSTEM] {self.prompt.system}")
        print(f"[HISTORY]\n{self.session.to_string()}[/HISTORY]")

        while True:
            user = input("User: ")
            if user == "q":
                break
            if user == "r":
                self.session.reset()
                continue

            self.session.add_question(user)
            prompt = self.prompt.prepare(session=self.session)

            x = mx.array([self.tokenizer.bos_token_id] + self.tokenizer.encode(prompt))  # type: ignore
            tokens = []
            skip = 0
            print("LLM: ", end="", flush=True)
            for _, token in enumerate(self.model.generate(x, self.temperature)):
                tokens.append(token.item())  # actually compute the token
                # checking if we reached max tokens
                if len(tokens) >= self.max_tokens:
                    break

                token_list = list(tokens)
                # checking if we reached end of sentence
                if token_list[-1] == self.tokenizer.eos_token_id:
                    break

                answer = self.tokenizer.decode(token_list)

                # checking if we reached end of sentence for LLM
                if answer_status(answer, self.prompt.TEXT_END) == STOP:
                    break
                if answer_status(answer, self.prompt.TEXT_END) == WAIT:
                    continue

                print(answer[skip:], end="", flush=True)
                skip = len(answer)

            self.session.add_answer(answer)
            if not answer.endswith("\n"):
                print("\n")



from typing import Callable, Dict, Optional, Tuple

from ._config import HFConfig, ModelConfig, QuantizeConfig
from .converter import llama_to_mlxllm, mistral_to_mlxllm, phi3_to_mlxllm
from .transformer import Transformer

MODEL_ENTRYPOINTS = {}


def register_model(name: Optional[str] = None) -> Callable:
    """Register a model entrypoint.

    Args:
        name (Optional[str], optional): mlx_llm model name. If None, is the same as the fn name. Defaults to None.

    Returns:
        Callable: model entrypoint
    """

    def wrapper(fn: Callable) -> Callable:
        key = name if name is not None else fn.__name__
        if key in MODEL_ENTRYPOINTS:
            raise ValueError(f"Model entrypoints already with '{key}' model.")
        MODEL_ENTRYPOINTS[key] = fn
        return fn

    return wrapper


@register_model("llama_2_7b_chat_hf")
def llama_2_7b_chat(
    vocab_size: int = 32000, norm_eps: float = 1e-5, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a LLaMA 2 7B chat model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 32000.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=4096,
        hidden_dim=11008,
        vocab_size=vocab_size,
        n_layers=32,
        n_heads=32,
        n_kv_heads=32,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(
        hf=HFConfig(
            repo_id="meta-llama/Llama-2-7b-chat-hf",
        ),
        converter=llama_to_mlxllm,
    )
    return model, config


@register_model("llama_2_7b_hf")
def llama_2_7b(
    vocab_size: int = 32000, norm_eps: float = 1e-5, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a LLaMA 2 7B chat model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 32000.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=4096,
        hidden_dim=11008,
        vocab_size=vocab_size,
        n_layers=32,
        n_heads=32,
        n_kv_heads=32,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(
        hf=HFConfig(
            repo_id="meta-llama/Llama-2-7b-hf",
        ),
        converter=llama_to_mlxllm,
    )
    return model, config


@register_model("llama_3_8b")
def llama_3_8b(
    vocab_size: int = 128256, norm_eps: float = 1e-5, rope_theta: float = 500000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a LLaMA 3 8B model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 128256.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (float, optional): rope theta. Defaults to 500000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=4096,
        hidden_dim=14336,
        vocab_size=vocab_size,
        n_layers=32,
        n_heads=32,
        n_kv_heads=8,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(
        hf=HFConfig(
            repo_id="meta-llama/Meta-Llama-3-8B",
        ),
        converter=llama_to_mlxllm,
    )
    return model, config


@register_model("llama_3_8b_instruct")
def llama_3_8b_instruct(
    vocab_size: int = 128256, norm_eps: float = 1e-5, rope_theta: float = 500000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a LLaMA 3 8B Instruct model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 128256.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (int, optional): rope theta. Defaults to 500000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """

    model = Transformer(
        dim=4096,
        hidden_dim=14336,
        vocab_size=vocab_size,
        n_layers=32,
        n_heads=32,
        n_kv_heads=8,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(
        hf=HFConfig(
            repo_id="meta-llama/Meta-Llama-3-8B-Instruct",
        ),
        converter=llama_to_mlxllm,
    )
    return model, config


@register_model("phi_3_mini_4k_instruct")
def phi3_mini_4k_instruct(
    vocab_size: int = 32064, norm_eps: float = 1e-5, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a Phi3 Mini 4k Instruct model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 32064.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=3072,
        hidden_dim=8192,
        vocab_size=vocab_size,
        n_layers=32,
        n_heads=32,
        n_kv_heads=32,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(
        hf=HFConfig(
            repo_id="microsoft/Phi-3-mini-4k-instruct",
        ),
        converter=phi3_to_mlxllm,
    )
    return model, config


@register_model("phi_3_mini_128k_instruct")
def phi3_mini_128k_instruct(
    vocab_size: int = 32064, norm_eps: float = 1e-5, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a Phi3 Mini 128k Instruct model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 32064.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=3072,
        hidden_dim=8192,
        vocab_size=vocab_size,
        n_layers=32,
        n_heads=32,
        n_kv_heads=32,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(
        hf=HFConfig(
            repo_id="microsoft/Phi-3-mini-128k-instruct",
        ),
        converter=phi3_to_mlxllm,
    )
    return model, config


@register_model("mistral_7b_instruct_v0.2")
def mistral_7b_instruct_v02(
    vocab_size: int = 32064, norm_eps: float = 1e-5, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a Mistral Instruct v0.2 Instruct model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 32000.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=4096,
        hidden_dim=14336,
        vocab_size=vocab_size,
        n_layers=32,
        n_heads=32,
        n_kv_heads=8,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(hf=HFConfig(repo_id="mistralai/Mistral-7B-Instruct-v0.2"), converter=mistral_to_mlxllm)

    return model, config


@register_model("openhermes_2.5_mistral_7b")
def openhermes_25_mistral_7b(
    vocab_size: int = 32064, norm_eps: float = 1e-5, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a OpenHermes 2.5 Mistral 7B model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 32000.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=4096,
        hidden_dim=14336,
        vocab_size=vocab_size,
        n_layers=32,
        n_heads=32,
        n_kv_heads=8,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(hf=HFConfig(repo_id="teknium/OpenHermes-2.5-Mistral-7B"), converter=mistral_to_mlxllm)

    return model, config


@register_model("tiny_llama_1.1B_chat_v1.0")
def tiny_llama_11B_chat_v10(
    vocab_size: int = 32000, norm_eps: float = 1e-5, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a Tiny Llama 1.1B Chat v1.0 model.

    Args:
        vocab_size (int, optional): vocab size. Defaults to 32000.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-5.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=2048,
        hidden_dim=5632,
        vocab_size=vocab_size,
        n_layers=22,
        n_heads=32,
        n_kv_heads=4,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
    )

    config = ModelConfig(hf=HFConfig(repo_id="TinyLlama/TinyLlama-1.1B-Chat-v1.0"), converter=llama_to_mlxllm)

    return model, config


@register_model("gemma_1.1_2b_it")
def gemma_2b_it(
    vocab_size: int = 256000, norm_eps: float = 1e-6, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a Gemma 2B (v1.1) model

    Args:
        vocab_size (int, optional): vocab size. Defaults to 256000.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-6.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=2048,
        hidden_dim=16384,
        vocab_size=vocab_size,
        n_layers=18,
        n_heads=8,
        n_kv_heads=1,
        head_dim=256,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
        gemma=True,
    )

    config = ModelConfig(hf=HFConfig(repo_id="google/gemma-1.1-2b-it"), converter=llama_to_mlxllm)

    return model, config


@register_model("gemma_1.1_7b_it")
def gemma_7b_it(
    vocab_size: int = 256000, norm_eps: float = 1e-6, rope_theta: float = 10000.0, rope_traditional: bool = False
) -> Tuple[Transformer, ModelConfig]:
    """Create a Gemma /B (v1.1) model

    Args:
        vocab_size (int, optional): vocab size. Defaults to 256000.
        norm_eps (float, optional): norm epsilon. Defaults to 1e-6.
        rope_theta (float, optional): rope theta. Defaults to 10000.0.
        rope_traditional (bool, optional): whether to use traditional rope. Defaults to False.

    Returns:
        Tuple[Transformer, ModelConfig]: model, config
    """
    model = Transformer(
        dim=3072,
        hidden_dim=24576,
        vocab_size=vocab_size,
        n_layers=28,
        n_heads=16,
        n_kv_heads=16,
        head_dim=256,
        norm_eps=norm_eps,
        rope_theta=rope_theta,
        rope_traditional=rope_traditional,
        gemma=True,
    )

    config = ModelConfig(hf=HFConfig(repo_id="google/gemma-1.1-7b-it"), converter=llama_to_mlxllm)

    return model, config



from ._base import Prompt
from .llama import LLaMA2Prompt, LLaMA3Prompt, TinyLLaMAPrompt
from .gemma import GemmaPrompt
from .mistral import MistralPrompt
from .phi import Phi3Prompt

PROMPT_ENTRYPOINTS = {
    "llama2": LLaMA2Prompt,
    "llama3": LLaMA3Prompt,
    "phi3": Phi3Prompt,
    "mistral": MistralPrompt,
    "tinyllama": TinyLLaMAPrompt,
    "gemma": GemmaPrompt,
}


def create_prompt(model_name: str, system: str = "") -> Prompt:
    """Create prompt based on model family

    Args:
        model_name (str): model name
        system (str): system prompt

    Returns:
        Prompt: prompt class
    """

    model_family = model_name.replace("_", "").replace(".", "").replace("-", "").lower()
    found = False
    for k in PROMPT_ENTRYPOINTS.keys():
        if k in model_family:
            model_family = k
            found = True
            break
    if not found:
        raise ValueError(
            f"Model {model_name} not found in available prompts. Familes available: {list(PROMPT_ENTRYPOINTS.keys())}"
        )

    return PROMPT_ENTRYPOINTS[model_family](system=system)  # type: ignore



from dataclasses import dataclass
from typing import Callable, Dict, Optional, Tuple


@dataclass
class HFConfig:
    repo_id: str
    revision: Optional[str] = None
    filename: Optional[str] = None


@dataclass
class QuantizeConfig:
    group_size: int
    bits: int


@dataclass
class ModelConfig:
    hf: HFConfig
    quantize: Optional[QuantizeConfig] = None
    converter: Optional[Callable] = None



from ..utils.session import Session
from .llama import TinyLLaMAPrompt


class Phi3Prompt(TinyLLaMAPrompt):
    """Phi3 Instruct prompt that follows this structure

    <|system|>
    You are a helpful AI assistant.<|end|>
    <|user|>
    How to explain Internet for a medieval knight?<|end|>
    <|assistant|>

    Args:
        system (str): system prompt

    """

    SYSTEM_START = "<|system|>"
    USER = "<|user|>"
    ASSISTANT = "<|assistant|>"
    TEXT_END = "<|end|>"

    def __init__(self, system: str) -> None:
        super().__init__(system)
